import numpy as np
from numpy import inf
import pandas as pd
from sklearn.tree import DecisionTreeClassifier
def get_tree_bins(data: pd.DataFrame,
                  col: str,
                  target: str = 'y',
                  max_depth=None,
                  max_leaf_nodes: int = 4,
                  min_samples_leaf: float = 0.05,
                  missing: list = [np.nan, None, ''],
                  precision: int = 2) -> list:
    """
    根据决策树对数据集进行特征分箱，默认最大分为5箱
    Args:
        data: DataFrame,将要根据决策树进行分箱的数据集
        col: str,将要分箱的列名
        target: str,目标变量名称，默认为'y'
        max_depth: int,树的深度
        max_leaf_nodes: int,最大叶子节点数,默认为 4
        min_samples_leaf: float,叶子节点样本数量最小占比,默认为0.05
        missing: list,缺失值列表
        precision:int,数据精度，小数点位数，默认为2

    Returns:
        根据决策树结果得到的分箱列表，例如：[-inf,30,40,50,inf]
    """
    df = data.copy()
    df[col].replace(missing, np.nan, inplace=True)

    data_notnull = df[[col, target]][df[col].notnull()]  # 删除空值
    result = []
    if len(np.unique(data_notnull[col])) > 1:
        x = data_notnull[col].values.reshape(-1, 1)
        y = data_notnull[target].values

        clf = DecisionTreeClassifier(criterion='entropy',  # “信息熵”最小化准则划分
                                     max_depth=max_depth,  # 树的深度
                                     max_leaf_nodes=max_leaf_nodes,  # 最大叶子节点数
                                     min_samples_leaf=min_samples_leaf)  # 叶子节点样本数量最小占比
        clf.fit(x, y)

        threshold = np.unique(clf.tree_.threshold)
        x_num = np.unique(x)

        for i in threshold:
            if i != -2:
                point = np.round(max(x_num[x_num < i]), precision)  # 取切分点左边的数
                result.extend([point])
        result = [float(str(i)) for i in list(np.unique(result))]
        result = [-inf] + result + [inf]
    return result



import numpy as np
from numpy import inf
import pandas as pd

from westat.logger import logger


def get_col_psi(data_actual: pd.DataFrame,
                data_expected: str,
                col: str = 'Score',
                bins: list = [],
                qcut=10,
                missing: list = [np.nan, None, ''],
                target='y',
                precision=2,
                language='en') -> pd.DataFrame:
    """
    计算数据集PSI
    Args:
        data_actual: DataFrame,实际数据集
        data_expected: DataFrame,预期数据集
        col:str,需要计算PSI的列名
        bins:list，计算PSI的分箱
        qcut: int,等额分箱的分组数
        missing: list,缺失值列表
        target: str,目标变量名称，默认为'y'
        precision:数据精度，小数点位数，默认为2
        language: str,数据结果标题列显示语言，默认为 'en',可手动修改为'cn'

    Returns:
        结果数据集保存数据集对应的PSI
    """
    logger.info('计算特征的PSI...')
    if len(bins) > 0:
        data_actual['Bins Range'] = pd.cut(data_actual[col], bins, duplicates='drop')
        data_expected['Bins Range'] = pd.cut(data_actual[col], bins, duplicates='drop')
    else:
        data_actual['Bins'] = data_actual[col].replace(missing, np.nan)
        data_expected['Bins'] = data_expected[col].replace(missing, np.nan)

        # 获取切分点
        bins_actual = pd.qcut(data_actual['Bins'], q=qcut, duplicates='drop', retbins=True)[1]
        bins_actual[0] = -inf
        bins_actual[-1] = inf

        # 根据切分点进行分组
        data_actual['Bins Range'] = pd.cut(data_actual['Bins'], bins=bins_actual)
        data_expected['Bins Range'] = pd.cut(data_expected['Bins'], bins=bins_actual)

    # 计算分组后的统计量
    df_actual = data_actual.groupby(by='Bins Range')['Bins Range'].count()
    df_expected = data_expected.groupby(by='Bins Range')['Bins Range'].count()
    df_expected.index = df_actual.index
    result = pd.concat([df_actual, df_expected], axis=1)
    result.columns = ['#Actual', '#Expected']

    # 计算统计量
    result['#Total'] = result['#Actual'] + result['#Expected']
    result['%Actual'] = result['#Actual'] / result['#Actual'].sum()
    result['%Expected'] = result['#Expected'] / result['#Expected'].sum()
    result['%Total'] = result['#Total'] / result['#Total'].sum()
    result['PSI'] = (result['%Actual'] - result['%Expected']) * np.log(result['%Actual'] / result['%Expected'])
    result['Total PSI'] = result['PSI'].fillna(0).sum()
    result['Bins Range'] = result.index
    result.reset_index(drop=True, inplace=True)
    result['No.'] = result.index

    # 设置显示格式
    result['No.'] = result['No.'] + 1
    result['Name'] = col
    result['#Total'] = result['#Total'].apply(lambda x: round(x, precision))
    result['#Actual'] = result['#Actual'].apply(lambda x: round(x, precision))
    result['#Expected'] = result['#Expected'].apply(lambda x: round(x, precision))
    result['%Total'] = result['%Total'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['%Actual'] = result['%Actual'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['%Expected'] = result['%Expected'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['PSI'] = result['PSI'].apply(lambda x: round(x, precision))
    result['Total PSI'] = result['Total PSI'].apply(lambda x: round(x, precision))

    # 标题栏语言设置
    if language == 'cn':
        result = result[
            ['No.', 'Name', 'Bins Range', '#Total', '#Actual', '#Expected', '%Total', '%Actual', '%Expected', 'PSI',
             'Total PSI']]
        result.rename(
            columns={'No.': '序号', 'Name': '名称', 'Bins Range': '分组', '#Total': '#合计', '#Actual': '#实际',
                     '#Expected': '#期望',
                     '%Total': '%合计', '%Actual': '%实际', '%Expected': '%期望', 'Total PSI': 'PSI 合计'},
            inplace=True)

    else:
        result = result[
            ['No.', 'Name', 'Bins Range', '#Total', '#Actual', '#Expected', '%Total', '%Actual', '%Expected', 'PSI',
             'Total PSI']]

    return result


def view_col_psi(data_actual: pd.DataFrame,
                 data_expected: str,
                 col: str = 'Score',
                 bins: list = [],
                 qcut=10,
                 missing: list = [np.nan, None, ''],
                 target='y',
                 color: str = '#007bff',
                 precision=2,
                 language='en') -> pd.DataFrame:
    """
    计算数据集PSI
    Args:
        data_actual: DataFrame,实际数据集
        data_expected: DataFrame,预期数据集
        col:str,需要计算PSI的列名
        bins:list，计算PSI的分箱
        qcut: int,等额分箱的分组数
        missing: list,缺失值列表
        target: str,目标变量名称，默认为'y'
        color:str,显示颜色
        precision:数据精度，小数点位数，默认为2
        language: str,数据结果标题列显示语言，默认为 'en',可手动修改为'cn'

    Returns:
        结果数据集保存数据集对应的PSI
    """
    result = get_col_psi(data_actual=data_actual,
                         data_expected=data_expected,
                         col=col,
                         bins=bins,
                         qcut=qcut,
                         missing=missing,
                         target=target,
                         precision=precision,
                         language=language)
    result['PSI.'] = result['PSI'].replace('', np.nan)
    result['PSI.'] = result['PSI.'].apply(lambda x: float(x))
    result = result.style.bar(subset=['PSI.'], color=color)
    return result



# 版本号
global version
version = '0.1.2'

def get_version():
    return version


import sys
import os
from datetime import datetime

import numpy as np
from numpy import inf
import pandas as pd
import matplotlib.pyplot as plt
from tqdm.notebook import tqdm

plt.rcParams['font.sans-serif'] = 'SimHei'  # 设置中文字体
plt.rcParams['axes.unicode_minus'] = False  # 用来正常显示负号

# 导入westat功能模块
# 日志模块
from westat.logger import logger

# 数据准备 sample
from pandas import read_csv, read_excel


# 数据探索 explorer
from westat.get_data_partition import get_data_partition
from westat.get_data_distribution import get_data_distribution
from westat.get_data_describe import get_data_describe


# 数据处理 modify
from westat.get_data_discrete import get_data_discrete
from westat.get_data_iv import get_data_iv
from westat.get_data_woe import get_data_woe
from westat.get_data_woe_transform import get_data_woe_transform
from westat.get_tree_bins import get_tree_bins
from westat.get_woe_iv import get_woe_iv, view_woe_iv

from westat.check_data_target import check_data_target
# from westat.dataframe_to_table import dataframe_to_table
from westat.get_col_type import get_col_type
from westat.get_col_bins import get_col_bins
from westat.get_col_psi import get_col_psi,view_col_psi
from westat.get_modify_bins import get_modify_bins
from westat.feature_selection import get_feature_by_ivcorr

from westat.get_model_iv import get_model_iv


# 模型开发 model
from westat.get_scorecard import get_scorecard
from westat.get_predict_score import get_predict_score


# 模型评估 access
from westat.get_score_distribution import get_score_distribution, view_score_distribution
from westat.get_auc import get_auc
from westat.get_ks import get_ks


# 绘图
from westat.plot_woe import plot_woe
from westat.plot_iv import plot_iv
from westat.plot_corr import plot_corr
from westat.plot_lift import plot_lift
from westat.plot_roc_ks import plot_roc_ks

# 决策树文件转换
from westat.tree_to_img import tree_to_img
from westat.tree_to_pdf import tree_to_pdf

# 日期处理函数
from westat.date_diff import date_diff

# 设置函数别名
get_data_desc = get_data_describe
get_data_dist = get_data_distribution
get_data_part = get_data_partition
get_score_dist = get_score_distribution
view_score_dist = view_score_distribution




class Table(pd.DataFrame):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return 'table for westat'



import numpy as np
import pandas as pd
from tqdm.notebook import tqdm
from westat.logger import logger
from westat.get_data_iv import get_data_iv
from westat.get_col_type import get_col_type


def get_feature_by_ivcorr(data: pd.DataFrame,
                          data_iv: pd.DataFrame = pd.DataFrame(),
                          min_iv: float = 0.02,
                          max_corr: float = 0.6,
                          keep: list = [],
                          drop: list = [],
                          target: str = 'y',
                          return_drop: bool = False,
                          precision:int = 2):
    """
    根据最小IV值 和 最大相关性 筛选特征
    Args:
        data:DataFrame,将要筛选特征的数据集
        data_iv:DataFrame,列的IV汇总，包含Name,IV 两列的数据集
        min_iv:筛选后允许的最小IV值
        max_corr:筛选后允许的最大相关性
        keep:list,需要保留的特征
        drop:list,需要删除的特征
        target:str,目标变量名称，默认为'y'
        return_drop:是否返回已删除的特征
        precision:数据精度，小数点位数，默认为2

    Returns:
        默认返回筛选后的特征名单，
        当return_drop为True时，同时返回筛选后的特征名单、删除的特征名单、相关性和IV数据表
    """
    df = data.copy()

    # 如果没有特征IV入参，则按照默认的决策树方法，批量计算特征IV
    if data_iv.empty:
        data_iv = get_data_iv(data, target=target)

    # 根据每一列的IV值，计算相关矩阵
    col_iv_filter = data_iv[data_iv['Name'].isin(df.columns)]

    # 如果列的数据类型不是连续型特征，则赋值为0后计算相关性，否则直接计算相关性
    col_type = get_col_type(data)
    for col in col_iv_filter['Name']:
        if col_type['Type'][col_type['Name'] == col].iloc[0] != 'continuous':
            df[col] = 0

    data_corr = df[col_iv_filter['Name']].corr()

    col_iv_result = []
    for i in range(len(col_iv_filter)):
        for j in range(len(col_iv_filter)):
            col1 = col_iv_filter.iloc[i, 0]
            iv1 = col_iv_filter.iloc[i, 1]
            col2 = col_iv_filter.iloc[j, 0]
            iv2 = col_iv_filter.iloc[j, 1]
            col_iv_result.append([col1, col2, iv1, iv2, iv1 - iv2])

    result = pd.DataFrame(col_iv_result, columns=['Name1', 'Name2', 'IV1', 'IV2', 'IV1-IV2'])
    result['Corr'] = data_corr.values.reshape(-1, 1)
    result = result[result['Name1'] != result['Name2']]

    col_drop_by_iv = result['Name1'][result['IV1'] <= min_iv].unique()
    col_drop_by_corr = result['Name1'][(result['Corr'] > max_corr) & (result['IV1-IV2'] < 0)].unique()

    col_result = [col for col in result['Name1'].unique() if
                  col not in col_drop_by_iv and col not in col_drop_by_corr and col not in drop]
    col_keep = sorted(list(set(col_result + keep)))

    # 设置显示格式
    result['IV1'] = result['IV1'].apply(lambda x: round(x, precision))
    result['IV2'] = result['IV2'].apply(lambda x: round(x, precision))
    result['IV1-IV2'] = result['IV1-IV2'].apply(lambda x: round(x, precision))
    result['Corr'] = result['Corr'].apply(lambda x: round(x, precision))


    # 是否返回已删除特征和相关矩阵
    if return_drop:
        return col_keep, col_drop_by_iv, col_drop_by_corr, result
    else:
        return col_keep


import pandas as pd
def check_data_target(data:pd.DataFrame) -> bool:
    """
    检查数据集中是否有列名为"y" 或 "target"的目标变量
    Args:
        data: 需要检查的数据集

    Returns:
        bool:返回一个布尔变量
        返回 True,表示数据集中有名为"y" 或 "target"的目标变量
        返回 False,表示数据集中有没有名为"y" 或 "target"的目标变量
    """
    col_list = [x.lower() for x in data.columns]
    if 'y' not in col_list and 'target' not in col_list:
        return False
    else:
        return True



import numpy as np
import pandas as pd
import statsmodels.api as sm

from westat.get_col_type import get_col_type
from westat.get_data_discrete import get_data_discrete
from westat.get_data_woe_transform import get_data_woe_transform
from westat.get_model_iv import get_model_iv
from westat.get_data_iv import get_data_iv
from westat.logger import logger


def get_scorecard(data: pd.DataFrame,
                  col_bins: pd.DataFrame = pd.DataFrame(),
                  col_dict: pd.DataFrame = pd.DataFrame(),
                  init_score: int = 600,
                  pdo: int = 20,
                  target: str = 'y',
                  return_lr: bool = False,
                  precision: int = 2,
                  language: str = 'en'):
    """
    按照提供的分箱、数据字典，以逻辑回归方法开发评分卡
    Args:
        data:pd.DataFrame,需要开发评分卡的目标数据集
        col_bins:pd.DataFrame,列的分箱
        col_dict:pd.DataFrame,列的数据字典
        init_score:初始模型分,默认为600
        pdo: int,坏件率每上升一倍，增加的分数，默认为20
        target: str,目标变量名称，默认为'y'
        return_lr:bool,是否返回逻辑回归模型的详细内容
        precision: :int,数据精度，小数点位数，默认为2
        language:str,返回结果的标题语言，默认为英文

    Returns:
        评分卡结果表，包含模型全部特征，分箱，截距，参数，得分
    """

    logger.info('开始配置评分卡...')

    # 数据类型
    col_types = get_col_type(data)

    # 数据离散化
    data_discrete = get_data_discrete(data=data, data_bins=col_bins, target=target)

    # WoE 转换
    data_woe = get_data_woe_transform(data_discrete, target=target)

    # 根据手动调整后的分箱，批量计算IV
    data_iv = get_data_iv(data_discrete, target=target, criterion='discrete')

    # 逻辑回归模型
    y = data_woe[target]
    x = data_woe[[col for col in data.columns if col != target]]
    x = sm.add_constant(x)

    lr = sm.Logit(y, x).fit()

    # 评分卡特征准备
    result = get_model_iv(data_discrete=data_discrete,
                          col_iv=data_iv,
                          col_bins=col_bins,
                          col_dict=col_dict,
                          col_types=col_types,
                          target=target,
                          precision=precision)

    # 评分卡模型的截距项
    result['Intercept'] = lr.params[0]
    result['Label'].fillna('', inplace=True)

    # 评分卡模型的参数
    df_coef = pd.DataFrame(data=lr.params, columns=['Coef'])
    df_coef['Name'] = df_coef.index
    result = result.merge(df_coef, how='left', on='Name')

    # 计算评分卡模型的特征得分
    odds = data[target].sum() / (data[target].count() - data[target].sum())
    b = pdo / np.log(2)
    a = init_score + b * np.log(odds)
    result['Score'] = -b * result['WoE'] * result['Coef']

    # 设置显示格式
    result.reset_index(drop=True, inplace=True)  # 重置索引
    result['Intercept'] = result['Intercept'].apply(lambda x: round(x, precision))
    result['Coef'] = result['Coef'].apply(lambda x: round(x, precision))
    result['Score'] = result['Score'].apply(lambda x: round(x, precision))

    logger.info('评分卡配置完成...')

    # 语言设置
    if language == 'cn':
        result = result[['No.', 'Name', 'Label', 'Type', 'Bins No.', 'Bin', 'Bins', '#Total', '#Bad',
                         '#Good', '%Total', '%Bad', '%Good', '%BadRate', 'WoE', 'IV', 'Total IV',
                         'WoE.', 'Style', 'Intercept', 'Coef', 'Score']]
        result.rename(
            columns={'No.': '序号', 'Name': '名称', 'Label': '描述', 'Type': '类型', 'Bins No.': '分箱序号',
                     'Bin': '分箱', 'Bins': '切分点', '#Total': '#合计', '#Bad': '#坏', '#Good': '好',
                     '%Total': '%合计',
                     '%Bad': '%坏', '%Good': '%好', '%BadRate': '%坏件率', 'Total IV': 'IV合计', 'Intercept': '截距',
                     'Coef': '系数',
                     'Score': '分数', 'Style': '样式'}, inplace=True)
    else:
        result = result[['No.', 'Name', 'Label', 'Type', 'Bins No.', 'Bin', 'Bins', '#Total', '#Bad',
                         '#Good', '%Total', '%Bad', '%Good', '%BadRate', 'WoE', 'IV', 'Total IV',
                         'WoE.', 'Style', 'Intercept', 'Coef', 'Score']]

    if return_lr:
        return result, lr, a, b
    else:
        return result



import numpy as np
import pandas as pd
from westat.get_predict_score import get_predict_score


def get_auc(data: pd.DataFrame,
            score_card: pd.DataFrame,
            init_score: int = 600,
            pdo: int = 20,
            target: str = 'y',
            return_data: bool = False,
            precision: int = 2):
    """
    根据评分卡内容，对目标数据集的评分结果计算auc
    Args:
        data:pd.DataFrame,目标数据集
        score_card:pd.DataFrame，评分卡规则表
        init_score:int,初始模型分,默认为600
        pdo:int,坏件率每上升一倍，增加的分数，默认为20
        target:tr,目标变量名称，默认为'y'
        return_data:是否返回结果数据
        precision:int,数据精度，小数点位数，默认为2

    Returns:
        计算roc_auc的值，或根据要求返回结果数据
    """
    from sklearn.metrics import roc_curve, auc

    data_score_proba = get_predict_score(data, score_card, init_score=init_score, pdo=pdo, target=target,
                                         precision=precision)
    fpr, tpr, thresholds = roc_curve(data['y'], data_score_proba['Proba'], drop_intermediate=False)
    result = pd.DataFrame([fpr, tpr, thresholds]).T
    result.columns = ['fpr', 'tpr', 'thresholds']

    auc = round(auc(tpr, tpr), precision)
    if return_data:
        return auc, result
    else:
        return auc



import numpy as np
from numpy import inf
import pandas as pd


def get_data_describe(data: pd.DataFrame,
                      data_dict: pd.DataFrame = pd.DataFrame(),
                      missing: list = [np.nan, None, ''],
                      key: list = ['#Sum', '#Min', '#Mean', '#Q1', '#Median', '#Q3', '#Max', '#Std', 'Top1', 'Top2',
                                   'Top3'],
                      precision: int = 2,
                      language: str = 'en') -> pd.DataFrame:
    """
    获取目标数据集的描述统计信息
    Args:
        data:DataFrame,需要进行描述统计分析的数据集
        data_dict:DataFrame,数据字典，包含列名（特征英文名Name）和描述(特征中文名 Label)两列的数据集
        missing:list,缺失值列表
        key:list,额外增加的统计量关键词,例如：['#Sum','#Min','#Q1','#Median','#Q3','#Max','#Mode','#Var','#Std','#Mean','#Kurt','#Skew','#StdMean','#Range','#Cv','#Sum_of_squares','Top1','Top2','Top3']
        precision:数据精度，小数点位数，默认为2
        language:str,数据结果标题列显示语言，默认为 'en',可手动修改为'cn'

    Returns:
        描述统计结果数据集
    """
    # 获取数据描述
    df = data.copy()

    result = []
    col_total = len(df)
    for col in df.columns:
        # 数据类型
        col_dtype = str(df[col].dtype)

        # 缺失值替换
        df[col].replace(missing, [np.nan] * len(missing), inplace=True)

        # 数据角色
        if col.lower() == 'y' or col.lower() == 'target':
            col_role = 'target'
        elif col.lower().find('_id') > 0:
            col_role = 'ID'
        else:
            col_role = 'feature'

        # 统计量计算
        col_n = len(df[df[col].notnull()])
        col_missing = len(df[df[col].isna()])
        col_unique = df[col].nunique()

        # 如果列是数值型，并且不是全为空，则计算统计量
        if col_dtype in ('int64', 'float64', 'float32') and len(df[df[col].notnull()]):
            col_sum = df[col].sum()
            col_min = df[col].min()
            col_max = df[col].max()
            col_mean = df[col].mean()
            col_median = df[col].median()
            col_mode = df[col].mode().iloc[0]
            col_std = df[col].std()
            col_var = df[col].var()
            col_sem = df[col].sem()
            col_skew = df[col].skew()
            col_kurt = df[col].kurt()
            col_q1 = df[col].quantile(0.25)
            col_q3 = df[col].quantile(0.75)
            col_range = col_max - col_min
            col_sum_of_squares = df[col].pow(2).sum()

            if col_mean != 0:
                col_cv = col_std / col_mean
            else:
                col_cv = np.nan
        else:
            col_sum = np.nan
            col_min = np.nan
            col_max = np.nan
            col_mean = np.nan
            col_median = np.nan
            col_mode = np.nan
            col_std = np.nan
            col_var = np.nan
            col_sem = np.nan
            col_skew = np.nan
            col_kurt = np.nan
            col_q1 = np.nan
            col_q3 = np.nan
            col_range = np.nan
            col_sum_of_squares = np.nan
            col_cv = np.nan

        if len(df[df[col].notnull()]) > 0:
            col_top_list = df[col].value_counts().index
            if len(col_top_list) >= 3:
                col_top3 = col_top_list[2]
                col_top2 = col_top_list[1]
                col_top1 = col_top_list[0]
            elif len(col_top_list) >= 2:
                col_top3 = np.nan
                col_top2 = col_top_list[1]
                col_top1 = col_top_list[0]
            elif len(col_top_list) >= 1:
                col_top3 = np.nan
                col_top2 = np.nan
                col_top1 = col_top_list[0]
            else:
                col_top3 = np.nan
                col_top2 = np.nan
                col_top1 = np.nan
        result.append(
            [col, '', col_dtype, col_role, col_total, col_n, col_n / col_total, col_missing, col_missing / col_total,
             col_unique, col_unique / col_total, col_sum,
             col_min, col_mean, col_q1, col_median, col_q3, col_max, col_range, col_mode, col_var, col_std, col_cv,
             col_sem, col_skew, col_kurt, col_sum_of_squares, col_top1, col_top2, col_top3])
    result = pd.DataFrame(result,
                          columns=['Name', 'Describe', 'Type', 'Role', '#Count', '#N', '%N', '#Missing', '%Missing',
                                   '#Unique', '%Unique',
                                   '#Sum', '#Min', '#Mean', '#Q1', '#Median', '#Q3', '#Max', '#Range', '#Mode', '#Var',
                                   '#Std', '#Cv', '#StdMean', '#Skew', '#Kurt', '#Sum_of_squares', 'Top1', 'Top2',
                                   'Top3'])

    # 设置显示格式
    result['#Count'] = result['#Count'].apply(lambda x: round(x, precision))
    result['#N'] = result['#N'].apply(lambda x: round(x, precision))
    result['#Missing'] = result['#Missing'].apply(lambda x: round(x, precision))
    result['#Unique'] = result['#Unique'].apply(lambda x: round(x, precision))
    result['#Sum'] = result['#Sum'].apply(lambda x: round(x, precision))
    result['#Min'] = result['#Min'].apply(lambda x: round(x, precision))
    result['#Mean'] = result['#Mean'].apply(lambda x: round(x, precision))
    result['#Q1'] = result['#Q1'].apply(lambda x: round(x, precision))
    result['#Median'] = result['#Median'].apply(lambda x: round(x, precision))
    result['#Q3'] = result['#Q3'].apply(lambda x: round(x, precision))
    result['#Max'] = result['#Max'].apply(lambda x: round(x, precision))
    result['#Range'] = result['#Range'].apply(lambda x: round(x, precision))
    result['#Mode'] = result['#Mode'].apply(lambda x: round(x, precision))
    result['#Var'] = result['#Var'].apply(lambda x: round(x, precision))
    result['#Std'] = result['#Std'].apply(lambda x: round(x, precision))
    result['#Sum_of_squares'] = result['#Sum_of_squares'].apply(lambda x: round(x, precision))
    result['#StdMean'] = result['#StdMean'].apply(lambda x: round(x, precision))
    result['#Skew'] = result['#Skew'].apply(lambda x: round(x, precision))
    result['#Kurt'] = result['#Kurt'].apply(lambda x: round(x, precision))
    result['%N'] = result['%N'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['%Missing'] = result['%Missing'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['%Unique'] = result['%Unique'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['#Cv'] = result['#Cv'].apply(lambda x: format(x, '.' + str(precision) + '%'))

    result.index.name = 'No.'
    result.reset_index(drop=True, inplace=True)
    if language == 'cn':
        col_list = ['Name', 'Describe', 'Type', 'Role', '#Count', '#N', '%N', '#Missing', '%Missing', '#Unique',
                    '%Unique'] + key
        result = result[col_list]
        result.rename(columns={'Name': '名称', 'Describe': '描述', 'Type': '类型', 'Role': '角色', '#Count': '#数量',
                               '#Missing': '#缺失值',
                               '%Missing': '%缺失值', '#Unique': '#唯一值', '%Unique': '%唯一值', '#Sum': '#合计',
                               '#Min': '#最小值',
                               '#Mean': '#均值', '#Median': '#中位数', '#Max': '#最大值', '#Mode': '#众数',
                               '#Var': '#方差', '#Std': '#标准差',
                               '#Skew': '#偏度', '#Kurt': '#峰度', '#Q1': '下四分位数', 'Q3': '上四分位数',
                               '#StdMean': '#标准误差', '#Range': '#极差', '#Cv': '#变异系数',
                               '#Sum_of_squares': '#平方和'},
                      inplace=True)
    else:
        col_list = ['Name', 'Describe', 'Type', 'Role', '#Count', '#N', '%N', '#Missing', '%Missing', '#Unique',
                    '%Unique'] + key
        result = result[col_list]

    if not data_dict.empty:
        result = result.merge(data_dict, on='Name', how='left')
        result['Describe'][result.iloc[:, -1].notnull()] = result.iloc[:, -1]
        result = result.iloc[:, :-1]

    return result




import logging

# 日志设置
logging.getLogger().setLevel(logging.INFO)
FORMAT = '%(asctime)s [%(levelname)s] %(process)d %(module)s %(message)s'
logging.basicConfig(format = FORMAT )

logger = logging.getLogger(__name__)


import numpy as np
import pandas as pd
from westat.get_score_distribution import get_score_distribution


def plot_lift(data: pd.DataFrame,
              col='Score',
              qcut: int = 10,
              color: list = ['#1f77b4', '#d62728'],
              marker: str = 'o',
              linewidth=2,
              linestyle='-',
              style='default',
              figsize=(15, 8),
              target: str = 'y',
              return_data: bool = False,
              precision: int = 2,
              language: str = 'en'):
    """
    绘制条形图，展示data数据集中目标变量col的WoE值，可用于检查WoE分布是否满足单调性要求
    Args:
        data: DataFrame,目标数据集
        col: str,需要计算WoE和IV的列名
        qcut: int,等额分箱的分组数
        color:list,条形图颜色名称,默认为['#1f77b4', '#d62728']
        marker:折线图标记样式，默认为o，即圆点
        target: str,目标变量名称，默认为'y'
        linewidth:折线图的线宽，默认为1
        linestyle:折线图中线的类型，默认为 '-'
        style:绘图的样式风格
        figsize:图片大小，默认为(15,8)
        target:str,目标变量名称
        return_data:bool,是否返回提升度数据
        precision: int,数据精度，小数点位数，默认为2
        language: str,数据结果标题列显示语言，默认为 'en',可手动修改为'cn'

    Returns:
        matplotlib条形图结果
    """
    import matplotlib.pyplot as plt
    plt.rcParams['font.sans-serif'] = 'SimHei'  # 设置中文字体
    plt.rcParams['axes.unicode_minus'] = False  # 用来正常显示负号

    result = get_score_distribution(data=data,
                                    col=col,
                                    qcut=qcut,
                                    target=target,
                                    precision=precision)

    x = [str(x) for x in result['Score Range']]
    lift = result['Lift']
    total = result['#Total']

    fig = plt.figure(figsize=figsize)
    ax1 = fig.add_subplot(1, 1, 1)
    ax1.bar(x, total, align='center', color=color[0],
            label='Total =  {}'.format(round(result['#Total'].sum(), precision)))
    ax1.set_title(col)
    ax1.set_xlabel(col)
    ax1.set_ylabel('Count')

    ax2 = ax1.twinx()
    ax2.plot(x, lift, marker=marker,
             color=color[1],
             linewidth=linewidth,
             linestyle=linestyle,
             label='Lift ')
    ax2.set_ylabel('Lift')
    fig.legend(loc=1)
    plt.style.use(style)

    # 设置数字标签
    for a, b in zip(x, total):
        if b < 0:
            p = b - total.max() / 50
        else:
            p = b + total.max() / 50
        ax1.text(a, p, b, ha='center', va='center', fontsize=12, color=color[0])

    # 设置数字标签
    for a, b in zip(x, lift):
        if b < 0:
            p = b - lift.max() / 50
        else:
            p = b + lift.max() / 50

        ax2.text(a, p, b, ha='center', va='bottom', fontsize=12, color=color[1])

    # 设置绘图显示语言
    if language == 'cn':
        fig.suptitle('Lift 分布', fontsize=14, fontweight='bold')
    else:
        fig.suptitle('Lift Distributions', fontsize=14, fontweight='bold')

    plt.show()

    # 返回数据结果
    if return_data:
        return result



def tree_to_img(file_name, path='', dtype='svg'):
    """
    转换决策树到图片

    Args:
        file_name: 决策树文件路径
        path: 目标图片的存放路径
        dtype: 图片格式

    Returns:
        决策树转换到图片的结果
    """
    # 修改字体为中文
    file_txt = ''
    with open(file_name, 'r', encoding='utf-8') as f:
        for line in f:
            line = line.replace('helvetica', 'SimHei')
            file_txt = file_txt + line

    with open(file_name, 'w', encoding='utf-8') as f:
        f.write(file_txt)

    # 转换决策树到图片
    if path:
        file_path = path
    else:
        file_path = os.path.splitext(file_name)[0] + '.' + dtype

    str_exec = 'dot -T' + dtype + ' ' + file_name + ' -o ' + file_path  # 决策树转图片命令
    print(str_exec)
    result = os.popen(str_exec)
    return result


import numpy as np
from numpy import inf
import pandas as pd

from westat.get_tree_bins import get_tree_bins


def get_woe_iv(data: pd.DataFrame,
               col: str,
               criterion: str = 'tree',
               bins: list = [],
               qcut: int = 0,
               missing: list = [np.nan, None, ''],
               target: str = 'y',
               precision: int = 2,
               language: str = 'en') -> pd.DataFrame:
    """
    计算数据集中指定列的WOE和IV值
    Args:
        data: DataFrame,目标数据集
        col: str,需要计算WoE和IV的列名
        criterion: str,分箱方法，默认为'tree'表示使用决策树分箱,当取值为 'discrete'时，表示数据集已经离散化，直接计算WoE和IV
        bins: list,手动指定的分箱列表
        qcut: int,等额分箱的分组数
        missing: list,缺失值列表
        target: str,目标变量名称，默认为'y'
        precision:数据精度，小数点位数，默认为2
        language: str,数据结果标题列显示语言，默认为 'en',可手动修改为'cn'

    Returns:
        结果数据集保存各分组名称和分组对应的WoE和IV值
    """
    df = data.copy()

    # 检查数据集中是否存在预定义的缺失值
    missing_list = []
    for m in missing:
        if m in data[col]:
            missing_list.append(m)

    # 自定义分箱
    if len(bins) > 0:
        if len(missing_list) > 0 or len(data[data[col].isnull()]) > 0:
            df[col].replace(missing, [np.nan] * len(missing), inplace=True)
            df[col] = pd.cut(df[col], bins)
            df[col] = df[col].cat.add_categories('missing')
            df[col].fillna('missing', inplace=True)
        else:
            df[col] = pd.cut(df[col], bins)

    # 等频分箱
    elif qcut > 0:
        bins = []

        if len(missing_list) > 0 or len(data[data[col].isnull()]) > 0:
            df[col].replace(missing, [np.nan] * len(missing), inplace=True)
            df[col] = pd.qcut(x=df[col], q=qcut, duplicates='drop')
            df[col] = df[col].cat.add_categories('missing')
            df[col].fillna('missing', inplace=True)
        else:
            df[col] = pd.qcut(x=df[col], q=qcut, duplicates='drop')

    # 决策树分箱
    elif criterion == 'tree':
        if len(missing_list) > 0 or len(data[data[col].isnull()]) > 0:
            df[col].replace(missing, [np.nan] * len(missing), inplace=True)
            bins = get_tree_bins(data=data, col=col, target=target, max_depth=None, max_leaf_nodes=4,
                                 min_samples_leaf=0.05,
                                 precision=precision)
            df[col] = pd.cut(df[col], bins)
            df[col] = df[col].cat.add_categories('missing')
            df[col].fillna('missing', inplace=True)
        else:
            bins = get_tree_bins(data=df, col=col, target=target, max_depth=None, max_leaf_nodes=4,
                                 min_samples_leaf=0.05,
                                 precision=2)
            df[col] = pd.cut(df[col], bins)

    result = df.groupby(col)[target].agg([('#Bad', lambda target: (target == 1).sum()),
                                          ('#Good', lambda target: (target == 0).sum()),
                                          ('#Total', 'count')]).reset_index()
    result['Name'] = col
    result['%Bad'] = result['#Bad'] / result['#Bad'].sum()
    result['%Good'] = result['#Good'] / result['#Good'].sum()
    result['%Total'] = result['#Total'] / result['#Total'].sum()
    result['%BadRate'] = result['#Bad'] / result['#Total']
    result['WoE'] = np.log(result['%Bad'] / result['%Good'])
    result['IV'] = (result['%Bad'] - result['%Good']) * result['WoE']
    result.replace([-inf, inf], [0, 0], inplace=True)
    result['Total IV'] = result['IV'].sum()
    result = result.rename(columns={col: 'Bin'})

    bins = result['Bin'].unique()

    for i in range(len(bins)):
        for j in range(len(bins)):
            if ',' in ''.join(map(str, bins)):
                m = float(str(bins[i]).split(',')[0].replace('(', '').replace('-inf', '-99999999998').replace('missing',
                                                                                                              '-99999999999').replace(
                    '+', ''))
                n = float(str(bins[j]).split(',')[0].replace('(', '').replace('-inf', '-99999999998').replace('missing',
                                                                                                              '-99999999999').replace(
                    '+', ''))
                if m < n:
                    bins[i], bins[j] = bins[j], bins[i]
            else:
                if str(bins[i]).replace('missing', '-99999999999') < str(bins[j]).replace('missing', '-99999999999'):
                    bins[i], bins[j] = bins[j], bins[i]

    df = pd.DataFrame({'Bin': bins, 'No.': range(len(bins))})
    result = result.merge(df, how='left', on='Bin')

    # 设置显示格式
    result['No.'] = result['No.'] + 1
    result['%Total'] = result['%Total'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['%Bad'] = result['%Bad'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['%Good'] = result['%Good'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['%BadRate'] = result['%BadRate'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['WoE'] = result['WoE'].apply(lambda x: round(x, precision))
    result['IV'] = result['IV'].apply(lambda x: round(x, precision))
    result['Total IV'] = result['Total IV'].apply(lambda x: round(x, precision))

    result.sort_values(by='No.', inplace=True)
    result.reset_index(drop=True, inplace=True)

    result.replace([np.nan, 'nan%'], ['', ''], inplace=True)
    if language == 'cn':
        result = result[
            ['Name', 'No.', 'Bin', '#Total', '#Bad', '#Good', '%Total', '%Bad', '%Good', '%BadRate', 'WoE', 'IV',
             'Total IV']]
        result.rename(
            columns={'Name': '名称', 'No.': '分组序号', 'Bin': '分组逻辑', '#Total': '#合计', '#Bad': '#坏', '#Good': '#好',
                     '%Total': '%合计', '%Bad': '%坏', '%Good': '%好', '%BadRate': '%坏件率', 'Total IV': 'IV合计'},
            inplace=True)

    else:
        result = result[
            ['Name', 'No.', 'Bin', '#Total', '#Bad', '#Good', '%Total', '%Bad', '%Good', '%BadRate', 'WoE', 'IV',
             'Total IV']]

    return result


def view_woe_iv(data,
                col:str,
                criterion:str='tree',
                bins:list=[],
                qcut:int=0,
                missing:list = [np.nan, None, ''],
                target:str='y',
                color:str='#007bff',
                precision:int=2,
                language:str='en'):
    """
    计算数据集中指定列的WOE和IV值，并以图形化的形式，对WoE进行展示
    Args:
        data: DataFrame,目标数据集
        col: str,需要计算WoE和IV的列名
        target: str,目标变量名称，默认为'y'
        criterion: 分箱方法，默认为决策树分箱
        bins: list,手动指定的分箱列表
        qcut: int,等额分箱的分组数
        missing: list,缺失值列表
        color:str,显示颜色
        precision:数据精度，小数点位数，默认为2
        language: str,数据结果标题列显示语言，默认为 'en',可手动修改为'cn'

    Returns:
        结果数据集保存各分组名称和分组对应的WoE和IV值
    """
    result = get_woe_iv(data=data,
                        col=col,
                        criterion=criterion,
                        bins=bins,
                        qcut=qcut,
                        missing=missing,
                        target=target,
                        precision=precision,
                        language=language)
    result['WoE.'] = result['WoE'].replace('', np.nan)
    result['WoE.'] = result['WoE.'].apply(lambda x: float(x))
    result = result.style.bar(subset=['WoE.'], color=color)
    return result



import numpy as np
import pandas as pd
from westat.logger import logger

def get_modify_bins(col_bins: pd.DataFrame,
                    modify_list: list):
    """
    根据手动调整后的分箱，修改自动化生成的分箱结果
    Args:
        col_bins: 自动化生成的分箱结果
        modify_list: 手动调整后的分箱

    Returns:
        返回合并后的分箱结果
        如果手动有调整分箱，则按照调整后的分箱执行，否则按照自动化分箱结果执行
    """
    col_bins_adjust = pd.DataFrame(modify_list, columns=['Name', 'NewBins'])

    result = col_bins.merge(col_bins_adjust, on='Name', how='left')
    result = result.rename(columns={'Bins': 'OldBins', 'NewBins': 'Bins'})
    result['Bins'][result['Bins'].isnull()] = result['OldBins']
    return result[['Name', 'Bins']]



import numpy as np
import pandas as pd
from westat.get_woe_iv import get_woe_iv


def plot_woe(data: pd.DataFrame,
             col: str,
             target: str = 'y',
             criterion: str = 'tree',
             bins: list = [],
             qcut: int = 0,
             missing: list = [np.nan, None, ''],
             color: list = ['#1f77b4', '#d62728'],
             marker: str = 'o',
             linewidth=2,
             linestyle='-',
             style='default',
             figsize: tuple = (8, 5),
             return_data: bool = False,
             precision: int = 2,
             language: str = 'en'):
    """
    绘制条形图，展示data数据集中目标变量col的WoE值，可用于检查WoE分布是否满足单调性要求
    Args:
        data: DataFrame,目标数据集
        col: str,需要计算WoE和IV的列名
        target: str,目标变量名称，默认为'y'
        criterion: 分箱方法，默认为决策树分箱
        bins: list,手动指定的分箱列表
        qcut: int,等额分箱的分组数
        missing: list,缺失值列表
        color:list,条形图颜色名称,默认为['#1f77b4', '#d62728']
        marker:折线图标记样式，默认为o，即圆点
        linewidth:折线图的线宽，默认为1
        linestyle:折线图中线的类型，默认为 '-'
        style:绘图的样式风格
        figsize:图片大小，默认为(8,5)
        return_data:bool,是否返回IV计算结果数据
        precision: int,数据精度，小数点位数，默认为2
        language: str,数据结果标题列显示语言，默认为 'en',可手动修改为'cn'

    Returns:
        matplotlib条形图结果
    """
    import matplotlib.pyplot as plt
    plt.rcParams['font.sans-serif'] = 'SimHei'  # 设置中文字体
    plt.rcParams['axes.unicode_minus'] = False  # 用来正常显示负号

    result = get_woe_iv(data=data, col=col, target=target, criterion=criterion, bins=bins, qcut=qcut, missing=missing,
                        precision=precision)
    x = [str(x) for x in result['Bin']]
    woe = result['WoE']
    total = result['#Total']
    total_woe = result['WoE'].sum()

    fig = plt.figure(figsize=figsize)
    ax1 = fig.add_subplot(1, 1, 1)
    ax1.bar(x, total, align='center', color=color[0],
            label='Total =  {}'.format(round(result['#Total'].sum(), precision)))
    ax1.set_title(col)
    ax1.set_xlabel(col)
    ax1.set_ylabel('Count')

    ax2 = ax1.twinx()
    ax2.plot(x, woe, marker=marker,
             color=color[1],
             linewidth=linewidth,
             linestyle=linestyle,
             label='WoE = {}'.format(round(total_woe, precision)))
    ax2.set_ylabel('WoE')
    fig.legend(loc=1)
    plt.style.use(style)

    # 设置数字标签
    for a, b in zip(x, total):
        if b < 0:
            p = b - total.max() / 50
        else:
            p = b + total.max() / 50
        ax1.text(a, p, b, ha='center', va='center', fontsize=12, color=color[0])

    # 设置数字标签
    for a, b in zip(x, woe):
        if b < 0:
            p = b - woe.max() / 50
        else:
            p = b + woe.max() / 50

        ax2.text(a, p, b, ha='center', va='bottom', fontsize=12, color=color[1])

    # 设置绘图显示语言
    if language == 'cn':
        fig.suptitle('WoE 分布', fontsize=14, fontweight='bold')
    else:
        fig.suptitle('WoE Distributions', fontsize=14, fontweight='bold')

    plt.show()

    # 返回数据结果
    if return_data:
        return result



import numpy as np
import pandas as pd
from westat.get_predict_score import get_predict_score


def plot_roc_ks(data: pd.DataFrame,
                score_card: pd.DataFrame,
                init_score: int = 600,
                pdo: int = 20,
                target: str = 'y',
                return_data: bool = False,
                precision: int = 2):
    """
    根据评分卡内容，对目标数据集的评分结果计算auc 和ks,并绘图
    Args:
        data:pd.DataFrame,目标数据集
        score_card:pd.DataFrame，评分卡规则表
        init_score:int,初始模型分,默认为600
        pdo:int,坏件率每上升一倍，增加的分数，默认为20
        target:tr,目标变量名称，默认为'y'
        return_data:是否返回结果数据
        precision:int,数据精度，小数点位数，默认为2

    Returns:
        绘图auc和ks，或根据要求返回结果数据
    """
    from sklearn.metrics import roc_curve, auc
    import matplotlib.pyplot as plt
    plt.rcParams['font.sans-serif'] = 'SimHei'  # 设置中文字体
    plt.rcParams['axes.unicode_minus'] = False  # 用来正常显示负号

    data_score_proba = get_predict_score(data, score_card, init_score=init_score, pdo=pdo, target=target,
                                         precision=precision)
    fpr, tpr, thresholds = roc_curve(data['y'], data_score_proba['Proba'], drop_intermediate=False)
    roc_auc = auc(tpr, tpr)
    plt.figure(figsize=(20, 10))

    # ROC曲线
    plt.subplot(121)
    plt.title('ROC')
    plt.plot(fpr, tpr, 'b', label='AUC = ' + str(round(roc_auc, precision)))
    plt.legend(loc='lower right', fontsize=16)
    plt.plot([0, 1], [0, 1], 'r--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    plt.xlabel('FPR', fontsize=20)
    plt.ylabel('TPR', fontsize=20)

    # KS曲线
    plt.subplot(122)
    pre = sorted(data_score_proba['Proba'], reverse=True)
    num = [i * int(len(pre) / 10) for i in range(10)]
    num = num + [(len(pre) - 1)]
    ks_thresholds = [max(thresholds[thresholds <= pre[i]]) for i in num]
    data_ks = pd.DataFrame([fpr, tpr, thresholds, tpr - fpr]).T
    data_ks.columns = ['fpr', 'tpr', 'thresholds', 'ks']
    data_ks = pd.merge(data_ks, pd.DataFrame(ks_thresholds, columns=['thresholds']), on='thresholds', how='inner')
    data_ks.reset_index(drop=True, inplace=True)
    data_ks['No.'] = data_ks.index + 1
    data_ks = data_ks[['No.', 'fpr', 'tpr', 'thresholds', 'ks']]

    ks = round(max(data_ks['ks']), precision)
    plt.title('KS')
    plt.plot(data_ks.index, data_ks['tpr'])
    plt.plot(data_ks.index, data_ks['fpr'])
    plt.plot(data_ks.index, data_ks['tpr'] - data_ks['fpr'], label='K-S = ' + str(round(ks, precision)))

    # 设置数字标签
    a = data_ks['No.'][data_ks['ks'] == data_ks['ks'].max()].iloc[0] - 1
    plt.plot([a, a], [0, ks], 'o--', color='red')

    plt.text(a, ks + 0.02, ks, ha='center', va='bottom', fontsize=12)
    plt.legend(loc='lower right', fontsize=16)
    plt.xlim([0, 10])
    plt.ylim([0.0, 1.0])
    plt.xlabel('% of Population', fontsize=20)
    plt.ylabel('% of total good / %Bad', fontsize=20)
    plt.show()
    if return_data:
        return data_ks



import numpy as np
import pandas as pd
from westat.logger import logger
from westat.get_tree_bins import get_tree_bins
from westat.get_col_type import get_col_type

def get_col_bins(data: pd.DataFrame,
                 target: str = 'y',
                 max_depth=None,
                 max_leaf_nodes: int = 4,
                 min_samples_leaf: float = 0.05,
                 missing: list = [np.nan, None, ''],
                 precision=2) -> pd.DataFrame:
    """
    将数据集中所有列，根据决策树进行分箱
    连续变量的分箱，按小于等于，大于切分，空值单独归位一类，例如:['age', [-inf, 22, 35, 50, 60, inf]]
    Args:
        data:DataFrame,需要分箱的数据集
        target:str,目标变量名称，默认为'y'
        max_depth:int,树的深度
        max_leaf_nodes:最大叶子节点数,默认为 4
        min_samples_leaf:叶子节点样本数量最小占比,默认为0.05
        missing: list,缺失值列表
        precision:int,数据精度，小数点位数，默认为2

    Returns:
        返回数据集，包含列名，分箱结果两列，分箱结果使用list保存
    """
    col_types = get_col_type(data)
    col_continuous_bins = []
    logger.info('连续变量决策树分箱进行中。。。')
    for i in range(len(col_types)):
        col = col_types.iloc[i, 0]
        col_type = col_types.iloc[i, 1]
        if col_type == 'continuous' and col in data.columns and col != target:
            point = get_tree_bins(data=data[[col, target]],
                                  col=col,
                                  target=target,
                                  max_depth=max_depth,
                                  max_leaf_nodes=max_leaf_nodes,
                                  min_samples_leaf=min_samples_leaf,
                                  missing=missing,
                                  precision=precision)
            if point:
                col_continuous_bins.append([col, point])
    result = pd.DataFrame(col_continuous_bins, columns=['Name', 'Bins'])
    return result



import pandas as pd


def plot_corr(data: pd.DataFrame,
              col_dict: pd.DataFrame = pd.DataFrame(),
              rotation: list = [0, 0],
              figsize: tuple = (10, 6)):
    """
    绘制目标数据集中指定特征的相关性表
    Args:
        data: 目标数据集
        col_dict: 特征的数据字典，包含Name,Label两列
        rotation:list,x和y坐标轴文字方向,默认为[0,0] 即x轴文字和y轴文字均为水平
        figsize:tuple,图片大小
    Returns:
        返回相关性图表
    """
    import matplotlib.pyplot as plt
    plt.rcParams['font.sans-serif'] = 'SimHei'  # 设置中文字体
    plt.rcParams['axes.unicode_minus'] = False  # 用来正常显示负号

    import seaborn as sns
    if col_dict.empty:
        corr = data.corr()  # 计算各变量的相关性系数
    else:
        result = data.copy()
        for col in result.columns:
            if col in col_dict['Name'].to_list():
                label = col_dict[col_dict['Name'] == col].iat[0, -1]
                result.rename(columns={col: label}, inplace=True)

        corr = result.corr()  # 计算各变量的相关性系数

    fig = plt.figure(figsize=figsize)
    ax = sns.heatmap(corr, annot=True, cbar=True)
    ax.set_xticklabels(corr.columns, rotation=rotation[0])
    ax.set_yticklabels(corr.columns, rotation=rotation[1])
    plt.show()



# 预测模型分数

import numpy as np
import pandas as pd

from westat.logger import logger
from westat.get_data_discrete import get_data_discrete


def get_predict_score(data:pd.DataFrame,
                      scorecard:pd.DataFrame,
                      init_score:int=600,
                      pdo:int=20,
                      target:str='y',
                      precision:int=2):
    """
    根据评分卡表，对目标数据集进行预测得分和概率
    Args:
        data: pd.Dataframe,目标数据集
        scorecard: pd.Dataframe,评分卡表
        init_score:int,初始模型分,默认为600
        pdo:int,坏件率每上升一倍，增加的分数，默认为20
        target:str,目标变量名称，默认为'y'
        precision:int,数据精度，小数点位数，默认为2

    Returns:
        返回预测结果表，包含模型特征原始值，模型特征得分，汇总评分卡得分，预测概率
    """
    logger.info('预测用户分数中...')

    # 复制数据
    df = data.copy()

    # 数据离散化
    col_result = scorecard['Name'].unique().tolist() + [target]
    col_bins = scorecard[['Name', 'Bins']][scorecard['Type'] == 'continuous'].drop_duplicates('Name')
    data_discrete = get_data_discrete(df[col_result], data_bins=col_bins, target=target, precision=precision)

    result = pd.DataFrame()
    col_score_list = []
    for col in scorecard['Name'].unique():
        col_score = col + '_Score'
        col_score_list.append(col_score)
        bins = scorecard['Bin'][scorecard['Name'] == col].tolist()
        score = scorecard['Score'][scorecard['Name'] == col].tolist()
        # 原样显示特征原始值
        result[col] = df[col]

        # 计算每个特征的得分
        result[col_score] = data_discrete[col].replace(bins, score)

        # 设置显示格式
        result[col_score] = result[col_score].apply(lambda x: round(x, precision))

    # 计算模型得分和预测概率
    odds = data[target].sum() / (data[target].count() - data[target].sum())

    b = pdo / np.log(2)
    a = init_score + b * np.log(odds)
    result[target] = df[target]
    result['Score'] = a - b * scorecard['Intercept'][0] + result[col_score_list].sum(axis=1)
    result['Proba'] = 1 - 1 / (1 + np.e ** ((result['Score'] - a) / -b))

    # 设置显示格式
    result['Score'] = result['Score'].apply(lambda x: round(x, precision))
    result['Proba'] = result['Proba'].apply(lambda x: round(x, precision))

    return result



import numpy as np
import pandas as pd
from tqdm.notebook import tqdm
from westat.logger import logger
from westat.get_woe_iv import get_woe_iv


def get_data_iv(data: pd.DataFrame,
                target='y',
                criterion='tree',
                bins=[],
                qcut=0,
                missing: list = [np.nan, None, ''],
                precision=2):
    """
    批量获取变量IV值
    Args:
        data: DataFrame,目标数据集
        target: str,目标变量名称，默认为'y'
        criterion: str,分箱方法，默认为决策树分箱
        bins: list,手动指定的分箱列表
        qcut: int,等额分箱的分组数
        missing: list,缺失值列表
        precision: 数据精度，小数点位数，默认为2

    Returns:
        返回包含特征名称，IV值两列的数据集
    """
    logger.info('IV值计算中。。。')
    col_iv = []
    for col in tqdm([i for i in data.columns if i != target]):
        col_woe_iv = get_woe_iv(data=data,
                                col=col,
                                target=target,
                                criterion=criterion,
                                bins=bins,
                                qcut=qcut,
                                missing=missing,
                                precision=precision)
        col_iv.append([col, float(col_woe_iv['Total IV'].loc[0])])
    result = pd.DataFrame(col_iv, columns=["Name", "IV"])
    result.sort_values(by='IV', ascending=False, inplace=True)
    logger.info('IV值计算完成！')
    return result



import numpy as np
from numpy import inf
import pandas as pd
from westat.logger import logger


def get_score_distribution(data: pd.DataFrame,
                           col: str = 'Score',
                           qcut: int = 10,
                           target: str = 'y',
                           precision: int = 2,
                           language: str = 'en'):
    """
    根据评分，查看好坏客户分布和提升度 lift
    Args:
        data: pd.DataFrame,目标数据集,需要包含评分、目标变量
        col: str,评分特征的名称，默认为'Score'
        qcut: int,等额分箱的分组数
        target: str,目标变量名称，默认为'y'
        precision: 数据精度，小数点位数，默认为2
        language: str,数据结果标题列显示语言，默认为 'en',可手动修改为'cn'
    Returns:
        评分的分布表，包含各分数区间的好坏客户数量，占比，提升度
    """
    logger.info("开始检查评分卡分布...")
    df = data.copy()
    df[col] = df[col].astype(int)

    # 获取分箱切分点，并修改下限为-inf,上限为inf
    bins = pd.qcut(x=df[col], q=qcut, duplicates='drop', retbins=True)[1]
    bins[0] = -inf
    bins[-1] = inf

    # 对得分进行分组
    df['Score Range'] = pd.cut(df[col], bins=bins)

    # 设置统计量
    result = df.groupby('Score Range')[target].agg([('#Bad', lambda target: (target == 1).sum()),
                                                    ('#Good', lambda target: (target == 0).sum()),
                                                    ('#Total', 'count')]).reset_index()

    result['%Bad'] = result['#Bad'] / result['#Bad'].sum()
    result['%Good'] = result['#Good'] / result['#Good'].sum()
    result['%Total'] = result['#Total'] / result['#Total'].sum()
    result['%BadRate'] = result['#Bad'] / result['#Total']
    result['%BadRate Random'] = result['#Bad'].sum() / result['#Total'].sum()
    result['%CumBad'] = result['#Bad'].cumsum() / result['#Total'].cumsum()
    result['%CumTotal'] = result['#Total'].cumsum() / result['#Total'].sum()
    result['Lift'] = result['%CumBad'] / result['%CumTotal']

    # 设置显示格式
    result.reset_index(drop=True, inplace=True)
    result['No.'] = result.index + 1
    result['%Total'] = result['%Total'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['%Bad'] = result['%Bad'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['%Good'] = result['%Good'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['%BadRate'] = result['%BadRate'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['%BadRate Random'] = result['%BadRate Random'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['%CumBad'] = result['%CumBad'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['%CumTotal'] = result['%CumTotal'].apply(lambda x: format(x, '.' + str(precision) + '%'))
    result['Lift'] = result['Lift'].apply(lambda x: round(x, precision))

    if language == 'cn':
        result = result[
            ['No.', 'Score Range', '#Total', '#Bad', '#Good', '%Total', '%Bad', '%Good', '%BadRate', '%BadRate Random',
             '%CumBad', '%CumTotal', 'Lift']]
        result.rename(
            columns={'No.': '序号', 'Score Range': '分数区间', '#Total': '#合计', '#Bad': '#坏', '#Good': '#好',
                     '%Total': '%合计', '%Bad': '%坏',
                     '%Good': '%好', '%BadRate': '%坏件率', '%BadRate Random': '%随机坏件率', '%CumBad': '%累计坏',
                     '%CumTotal': '%累计合计', 'Lift': '提升度'}, inplace=True)
    else:
        result = result[
            ['No.', 'Score Range', '#Total', '#Bad', '#Good', '%Total', '%Bad', '%Good', '%BadRate', '%BadRate Random',
             '%CumBad', '%CumTotal', 'Lift']]
    return result


def view_score_distribution(data: pd.DataFrame,
                            col: str = 'Score',
                            qcut: int = 10,
                            target: str = 'y',
                            color: str = '#007bff',
                            precision: int = 2,
                            language: str = 'en'):
    """
        根据评分，查看好坏客户分布和提升度 lift
        Args:
            data: pd.DataFrame,目标数据集,需要包含评分、目标变量
            col: str,评分特征的名称，默认为'Score'
            qcut: int,等额分箱的分组数
            target: str,目标变量名称，默认为'y'
            color:str,显示颜色
            precision: 数据精度，小数点位数，默认为2
            language: str,数据结果标题列显示语言，默认为 'en',可手动修改为'cn'
        Returns:
            评分的分布表，包含各分数区间的好坏客户数量，占比，提升度
        """
    result = get_score_distribution(data=data, col=col, qcut=qcut, target=target, precision=precision,
                                    language=language)
    result['Lift.'] = result['Lift']
    result = result.style.bar(subset=['Lift.'], color=color)
    return result



import numpy as np
import pandas as pd
from westat.get_predict_score import get_predict_score


def get_ks(data: pd.DataFrame,
           score_card: pd.DataFrame,
           init_score: int = 600,
           pdo: int = 20,
           target: str = 'y',
           return_data: bool = False,
           precision: int = 2):
    """
    根据评分卡内容，对目标数据集的评分结果计算ks
    Args:
        data:pd.DataFrame,目标数据集
        score_card:pd.DataFrame，评分卡规则表
        init_score:int,初始模型分,默认为600
        pdo:int,坏件率每上升一倍，增加的分数，默认为20
        target:tr,目标变量名称，默认为'y'
        return_data:是否返回结果数据
        precision:int,数据精度，小数点位数，默认为2

    Returns:
        计算ks的值，或根据要求返回结果数据
    """
    from sklearn.metrics import roc_curve

    data_score = get_predict_score(data, score_card, init_score=init_score, pdo=pdo, target=target, precision=precision)
    fpr, tpr, thresholds = roc_curve(data['y'], data_score['Proba'], drop_intermediate=False)

    pre = sorted(data_score['Proba'], reverse=True)
    num = [i * int(len(pre) / 10) for i in range(10)]
    num = num + [(len(pre) - 1)]
    ks_thresholds = [max(thresholds[thresholds <= pre[i]]) for i in num]
    result = pd.DataFrame([fpr, tpr, thresholds, tpr - fpr]).T
    result.columns = ['fpr', 'tpr', 'thresholds', 'ks']
    result = pd.merge(result, pd.DataFrame(ks_thresholds, columns=['thresholds']), on='thresholds', how='inner')
    result.reset_index(drop=True, inplace=True)
    result['No.'] = result.index + 1
    result = result[['No.', 'fpr', 'tpr', 'thresholds', 'ks']]

    total = data_score.groupby(['Proba'])[target].count()
    bad = data_score.groupby(['Proba'])[target].sum()
    data_ks = pd.DataFrame({'#Total': total, '#Bad': bad})
    data_ks['#Good'] = data_ks['#Total'] - data_ks['#Bad']
    data_ks['Proba'] = data_ks.index
    data_ks.index = range(len(data_ks))
    data_ks = data_ks.sort_values(by='Proba', ascending=True)
    data_ks['%CumBad'] = data_ks['#Bad'].cumsum() / data_ks['#Bad'].sum()
    data_ks['%CumGood'] = data_ks['#Good'].cumsum() / data_ks['#Good'].sum()
    data_ks['KS'] = data_ks['%CumGood'] - data_ks['%CumBad']
    data_ks.reset_index(drop=True, inplace=True)
    data_ks['No.'] = data_ks.index + 1
    data_ks = data_ks[['No.', 'Proba', '#Total', '#Bad', '#Good', '%CumBad', '%CumGood', 'KS']]
    ks = round(result['ks'].max(), precision)

    if return_data:
        return ks, result
    else:
        return ks



import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

from westat.logger import logger


def get_data_partition(data: pd.DataFrame,
                       test_size: float = 0.25,
                       random_state=1234,
                       target: str = 'y'):
    """
        划分训练集和测试集
    Args:
        data: DataFrame,将要进行数据分区的目标数据集
        target: str,目标变量名称，默认为'y'
        test_size:float 测试集所占比例，默认为0.25
        random_state: 随机种子，默认为1234

    Returns:
        返回数据分区后的训练集和测试集
    """
    X = data[[col for col in data.columns if col != target]]
    y = data[target]
    x_train, x_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=random_state)
    data_train = x_train.reset_index()
    del data_train['index']
    data_train[target] = y_train.reset_index()[target]
    data_test = x_test.reset_index()
    del data_test['index']
    data_test[target] = y_test.reset_index()[target]
    return data_train, data_test



from prettytable import PrettyTable
from westat.logger import logger


def dataframe_to_table(data):
    """
    将pandas DataFrame 转换为 PrettyTable
    Args:
        data:将要转换的DataFrame

    Returns:

    """
    if data.empty:
        logger.warning('输入的数据集为空')
        return
    else:
        table = PrettyTable(list(data.columns))
        for row in range(len(data)):
            table.add_row(list(data.iloc[row,:]))
        print(table)
        return



import pandas as pd
from westat.logger import logger


def get_col_type(data: pd.DataFrame) -> pd.DataFrame:
    """
    划分离散和连续变量 （连续变量：int64,float64，float32；离散变量：其他）
    Args:
        data:需要划分离散和连续变量的数据集

    Returns:
        返回 DataFrame ,存放列名和列的类型（连续、离散）
    """
    logger.info('连续和离散变量划分中。。。')
    col_all = data.columns
    col_all_type = data.dtypes
    col_type = []
    for i in range(len(col_all)):
        if col_all[i] != 'y':
            if str(col_all_type[i]) in ('int64', 'float64', 'float32'):
                col_type.append([col_all[i], 'continuous'])
            else:
                col_type.append([col_all[i], 'discrete'])
    df = pd.DataFrame(col_type, columns=['Name', 'Type'])
    logger.info('连续和离散变量划分完成！')
    return df



import pandas as pd
from westat import tqdm
from westat.get_woe_iv import get_woe_iv


def get_model_iv(data_discrete: pd.DataFrame,
                 col_iv: pd.DataFrame,
                 col_bins: pd.DataFrame,
                 col_types: pd.DataFrame,
                 col_dict: pd.DataFrame = pd.DataFrame(),
                 target: str = 'y',
                 precision: int = 2,
                 language: str = 'en'):
    """
    根据给定的数据集，获取全部特征在给定分箱内的woe、iv
    Args:
        data_discrete:pd.DataFrame,已经离散化的数据集
        col_iv:pd.DataFrame,数据集所有特征的iv
        col_bins:pd.DataFrame,数据集所有特征的切分点
        col_types:pd.DataFrame,数据集所有特征的数据类型
        col_dict:pd.DataFrame,数据集所有特征的数据字典,默认为空
        target:str,目标变量名称，默认为'y'
        precision:int:数据精度，小数点位数，默认为2
        language:str,数据结果标题列显示语言，默认为 'en',可手动修改为'cn'

    Returns:
        根据已经离散化的数据集，返回特征在给定数据集的woe、iv
    """
    result = pd.DataFrame()
    for col in tqdm(data_discrete.columns):
        if col != target and col_iv['IV'][col_iv['Name'] == col].values[0] != 0:
            df_woe = get_woe_iv(data=data_discrete, col=col, criterion='discrete', target=target, precision=precision)
            result = pd.concat([result, df_woe])

    # 如果数据字典不为空，则填充字段说明，否则字段说明为空
    if col_dict.empty:
        result_1 = result
        result_1['Label'] = ''
    else:
        result_1 = result.merge(col_dict, on='Name', how='left')
        result_1['Label'] = result_1.iloc[:, -1]

    # 合并数据类型
    result_2 = result_1.merge(col_types, on='Name', how='left')

    # 合并数据分箱
    result = result_2.merge(col_bins, on='Name', how='left')
    result.sort_values(by=['Total IV', 'Name', 'No.'], ascending=[False, True, True], inplace=True)
    result.reset_index(drop=True, inplace=True)

    result['WoE.'] = result['WoE']
    result['Style'] = result['Name'].apply(lambda x: list(result['Name'].unique()).index(x) % 2)
    result['Label'].fillna('', inplace=True)

    # 特征的顺序
    result['Bins No.'] = result['No.']
    result['No.'] = result['Name'].apply(lambda x: list(result['Name'].unique()).index(x) + 1)

    # 语言设置
    if language == 'cn':
        result = result[['No.', 'Name', 'Label', 'Type', 'Bins No.', 'Bin', 'Bins', '#Total', '#Bad', '#Good', '%Total',
                         '%Bad', '%Good', '%BadRate', 'WoE', 'IV', 'Total IV', 'WoE.', 'Style']]
        result.rename(
            columns={'No.': '序号', 'Name': '名称', 'Label': '描述', 'Type': '类型', 'Bins No.': '分箱序号',
                     'Bin': '分箱', 'Bins': '切分点', '#Total': '#合计', '#Bad': '#坏', '#Good': '好',
                     '%Total': '%合计', '%Bad': '%坏', '%Good': '%好',
                     '%BadRate': '%坏件率', 'Total IV': 'IV合计', 'Style': '样式'}, inplace=True)
    else:
        result = result[['No.', 'Name', 'Label', 'Type', 'Bins No.', 'Bin', 'Bins', '#Total', '#Bad', '#Good', '%Total',
                         '%Bad', '%Good', '%BadRate', 'WoE', 'IV', 'Total IV', 'WoE.', 'Style']]
    return result



def tree_to_pdf(file_name, path=''):
    """
    转换决策树到pdf文件

    Args:
        file_name: 决策树文件路径
        path: 目标pdf文件的存放路径

    Returns:
        决策树转换到pdf文件的结果
    """
    # 修改字体为中文
    file_txt = ''
    with open(file_name, 'r', encoding='utf-8') as f:
        for line in f:
            line = line.replace('helvetica', 'SimHei')
            file_txt = file_txt + line

    with open(file_name, 'w', encoding='utf-8') as f:
        f.write(file_txt)

    # 转换决策树到图片 pdf文件
    if path:
        file_path = path
    else:
        file_path = os.path.splitext(file_name)[0] + '.pdf'

    str_exec = 'dot -Tpdf ' + file_name + ' -o ' + file_path  # 决策树转pdf命令
    print(str_exec)
    result = os.popen(str_exec)
    return result



import numpy as np
import pandas as pd
from tqdm.notebook import tqdm
from westat.get_col_bins import get_col_bins
from westat.logger import logger


def get_data_discrete(data: pd.DataFrame,                      
                      data_bins: pd.DataFrame = pd.DataFrame(),
                      max_depth=None,
                      max_leaf_nodes: int = 4,
                      min_samples_leaf: float = 0.05,
                      missing: list = [np.nan, None, ''],
                      target: str = 'y',
                      precision: int = 2):
    """
    将原始数据集，按照分箱数据集中的要求，进行离散化处理，并返回离散化后的数据集
    如果数据集中的列不在分箱数据集中，则保留原始数据集中的格式
    Args:
        data: DataFrame,需要进行离散化的目标数据集
        data_bins: DataFrame,存放列名，分箱列表的数据集
        max_depth: int,树的深度
        max_leaf_nodes:最大叶子节点数,默认为 4
        min_samples_leaf:float,叶子节点样本数量最小占比,默认为0.05
        missing: list,缺失值列表
        target:str,目标变量名称，默认为'y'
        precision:int,数据精度，小数点位数，默认为2

    Returns:
        返回离散化后的数据集
    """
    logger.info('根据分箱离散化连续变量进行中。。。')

    if data_bins.empty:
        bins = get_col_bins(data=data,
                            target=target,
                            max_depth=max_depth,
                            max_leaf_nodes=max_leaf_nodes,
                            min_samples_leaf=min_samples_leaf,
                            missing=missing,
                            precision=precision)
    else:
        bins = data_bins

    # 不在分箱清单中的列，保持原样
    cols = [i for i in data.columns if i not in [i[0] for i in bins]]
    result = data[cols].copy()

    # 其他列按照分箱清单进行离散化
    for i in tqdm(range(len(bins))):
        col = bins.iloc[i, 0]
        cut_points = bins.iloc[i, 1]
        if col in data.columns:
            result[col] = pd.cut(data[col], cut_points).astype('str')

    # 缺失值替换
    result.replace(missing, ['missing'] * len(missing), inplace=True)
    logger.info('根据分箱离散化连续变量完成！')
    return result



import pandas as pd


def get_data_distribution(data: pd.DataFrame,
                          col='y',
                          by=['#Count', 'Name'],
                          ascending=[True, False],
                          total=True,
                          display=False,
                          precision=2,
                          language='en') -> pd.DataFrame:
    """
    查看数据集data中col列的数据分布情况
    Args:
        data:DataFrame,需要检查数据分布的数据集
        by:list,根据指定列进行排序，默认根据数量（降序）、名称（升序）排序
        ascending:list,指定列对应的排序列表
        total:bool,是否显示合计信息
        col:需要检查数据分布的列，默认为'y'
        display:bool,是否显示详细信息
        precision:数据精度，默认为2位小数
        language:显示语言，en：显示为英文，cn：显示为中文
    Returns:
        返回包含数据量、数据占比的数据集
    """
    df_ratio = pd.DataFrame(data=data[col].value_counts())
    df_ratio['%Ratio'] = df_ratio[col] / len(data)
    df_ratio['Value'] = df_ratio.index
    df_ratio['Name'] = col
    df_ratio = df_ratio.rename(columns={col: '#Count'})
    df_ratio = df_ratio[['Name', 'Value', '#Count', '%Ratio']]

    # 排序
    df_ratio.sort_values(by=by, ascending=ascending, inplace=True)

    # 拼接汇总结果
    df_ratio = pd.concat(
        [df_ratio, pd.DataFrame({'Name': '', 'Value': '', '#Count': df_ratio['#Count'].sum(), '%Ratio': 1}, index=[0])])
    df_ratio.reset_index(drop=True, inplace=True)
    df_ratio['No.'] = df_ratio.index + 1

    # 数据精度设置
    df_ratio['%Ratio'] = df_ratio['%Ratio'].apply(lambda x: format(x, '.' + str(precision) + '%'))

    # 标题栏语言设置
    if language == 'cn':
        df_ratio.rename(columns={'No.': '序号', 'Name': '名称', 'Value': '值', '#Count': '#数量', '%Ratio': '%占比'},
                        inplace=True)
        if display:
            df_ratio = df_ratio[['序号', '名称', '值', '#数量', '%占比']]
            # 是否显示合计
            if total:
                df_ratio.iloc[-1, 1] = '合计'
            else:
                df_ratio = df_ratio[df_ratio['名称'] != '']
        else:
            df_ratio = df_ratio[['值', '#数量', '%占比']]
            # 是否显示合计
            if total:
                df_ratio.iloc[-1, 0] = '合计'
            else:
                df_ratio = df_ratio[df_ratio['名称'] != '']

    else:
        if display:
            df_ratio = df_ratio[['No.', 'Name', 'Value', '#Count', '%Ratio']]
            # 是否显示合计
            if total:
                df_ratio.iloc[-1, 1] = 'Total'
            else:
                df_ratio = df_ratio[df_ratio['Value'] != '']
        else:
            df_ratio = df_ratio[['Value', '#Count', '%Ratio']]
            # 是否显示合计
            if total:
                df_ratio.iloc[-1, 0] = 'Total'
            else:
                df_ratio = df_ratio[df_ratio['Value'] != '']

    return df_ratio



from datetime import datetime
from westat.logger import logger

def date_diff(start_date:str,
              end_date:str,
              unit:str='d'):
    """
    根据输入的开始日期和结束日期，计算日期差、月份差、年份差，并保留2位小数
    Args:
        start_date: str,开始日期
        end_date: str,结束日期
        unit: str,计算类型，d-日期差，m-月份差，y-年份差，默认为日期差

    Returns:
        int/float,计算日期差、月份差、年份差,计算日期差时返回int，否则返回float
    """
    if len(start_date) < 10 or len(end_date) < 10:
        date1 = datetime.strptime(start_date[0:7], '%Y-%m')
        date2 = datetime.strptime(end_date[0:7], '%Y-%m')
    else:
        date1 = datetime.strptime(start_date, '%Y-%m-%d')
        date2 = datetime.strptime(end_date, '%Y-%m-%d')

    if unit == 'd':
        return int((date2 - date1).days)
    elif unit == 'm':
        from dateutil.relativedelta import relativedelta
        delta = relativedelta(date2, date1)
        return round(delta.years * 12 + delta.months + delta.days / 30, 2)
    elif unit == 'y':
        from dateutil.relativedelta import relativedelta
        delta = relativedelta(date2, date1)
        return round(delta.years + delta.months / 12 + delta.days / 365, 2)
    else:
        logger.warning('输入错误， unit：{} 不支持'.format(unit))


import numpy as np
import pandas as pd
from tqdm.notebook import tqdm
from westat.logger import logger
from westat.get_woe_iv import get_woe_iv


def get_data_woe_transform(data_discrete,
                           target='y',
                           criterion='discrete',
                           missing: list = [np.nan, None, ''],
                           precision=2):
    """
    根据离散化数据集，进行WoE转换
    Args:
        data_discrete: DataFrame,目标数据集
        target: str,目标变量名称，默认为'y'
        criterion: str,分箱方法，默认为'discrete'，表示使用已经离散化的数据计算WoE
        missing: list,缺失值列表
        precision: 数据精度，小数点位数，默认为2

    Returns:
        返回经过WoE转换后的数据集
    """
    logger.info('WOE转换中。。。')
    data_woe = pd.DataFrame()
    for col in tqdm([i for i in data_discrete.columns if i != target]):
        col_woe = get_woe_iv(data_discrete, col=col, target=target, criterion=criterion, missing=missing, precision=precision)
        s = data_discrete[col].replace(list(col_woe['Bin']), list(col_woe['WoE']))
        data_woe = pd.concat([data_woe, s], axis=1)
    data_woe[target] = data_discrete[target]
    logger.info('WOE转换完成！')
    return data_woe
